import pyenv
import sys
import random


def run_test(
    pyenv_build,
    pyproject_toml,
    test_package,
    prerequisites,
    strategy_name,
    _storybook,
    _doctests,
):
    if strategy_name == "latest":
        strategies = [
            lambda versions: versions[-1],
        ]
    elif strategy_name == "earliest":
        strategies = [
            lambda versions: versions[0],
        ]
    elif strategy_name in ("full", "maxi"):
        strategies = (
            [
                lambda versions: versions[0],
            ]
            + [random.choice] * 2
            if strategy_name == "full"
            else 5 + [lambda versions: versions[-1]]
        )
    else:
        raise Exception(f"Strategy name {strategy_name} not found")

    for strategy in strategies:
        envirotestvenv = pyenv.EnvirotestVirtualenv(
            pyenv_build=pyenv_build,
            pyproject_toml=pyproject_toml,
            picker=strategy,
            test_package=test_package,
            prerequisites=prerequisites,
        )
        envirotestvenv.build()

        python_path = envirotestvenv.venv.python_path

        _doctests(python_path)
        results = (
            _storybook(python_path=python_path)
            .only_uninherited()
            .ordered_by_name()
            .play()
        )
        if not results.all_passed:
            print("FAILED")
            print("COPY the following into hitch/devenv.yml:\n\n")
            print("python version: {}".format(envirotestvenv.python_version))
            print("packages:")
            for package, version in envirotestvenv.picked_versions.items():
                print("  {}: {}".format(package, version))
            sys.exit(1)



from hitchrunpy import HitchRunPyException
from hitchstory import BaseEngine, GivenDefinition, GivenProperty
from hitchstory import no_stacktrace_for, validate
from strictyaml import Map, MapPattern, Optional, Str, Int
from templex import Templex
import time

"""
class Engine(BaseEngine):
    given_definition = GivenDefinition(
        python_version=GivenProperty(Str()),
        setup=GivenProperty(Str()),
        files=GivenProperty(
            MapPattern(Str(), Str()),
            inherit_via=GivenProperty.OVERRIDE,
        ),
    )

    def __init__(self, build, rewrite=False, cprofile=False):
        self._build = build
        self._rewrite = rewrite
        self._cprofile = cprofile

    def set_up(self):
        self._build.ensure_built()

        for filename, contents in self.given.get("files", {}).items():
            filepath = self._build.working.parent.joinpath(filename)
            if not filepath.dirname().exists():
                filepath.dirname().makedirs()
            filepath.write_text(contents)

    def assert_text(self, will_output, actual_output):
        Templex(will_output).assert_match(actual_output)

    def _story_friendly_output(self, text):
        return text

    @no_stacktrace_for(AssertionError)
    @no_stacktrace_for(HitchRunPyException)
    @validate(
        code=Str(),
        will_output=Str(),
        environment_vars=MapPattern(Str(), Str()),
        raises=Map({Optional("type"): Str(), Optional("message"): Str()}),
    )
    def run(self, code, will_output=None, environment_vars=None, raises=None):
        self.example_py_code = (
            self._build.example_python_code.with_terminal_size(160, 100)
            .with_env(**{} if environment_vars is None else environment_vars)
            .with_setup_code(self.given.get("setup", ""))
        )

        to_run = self.example_py_code.with_code(code)

        if self._cprofile:
            to_run = to_run.with_cprofile(
                self.path.profile.joinpath("{0}.dat".format(self.story.slug))
            )

        result = (
            to_run.expect_exceptions().run() if raises is not None else to_run.run()
        )

        actual_output = self._story_friendly_output(result.output)

        if will_output is not None:
            try:
                self.assert_text(will_output, actual_output)
            except AssertionError:
                if self._rewrite:
                    self.current_step.update(**{"will output": actual_output})
                else:
                    raise

        if raises is not None:
            exception_type = raises.get("type")
            message = raises.get("message")

            try:
                result.exception_was_raised(exception_type)
                exception_message = self._story_friendly_output(
                    result.exception.message
                )
                self.assert_text(exception_message, message)
            except AssertionError:
                if self._rewrite:
                    new_raises = raises.copy()
                    new_raises["message"] = exception_message
                    self.current_step.update(raises=new_raises)
                else:
                    raise

    @validate(seconds=Int())
    def sleep(self, seconds):
        time.sleep(int(seconds))

    def on_success(self):
        if self._rewrite:
            self.new_story.save()
"""



import dirtemplate
from strictyaml import load
from git import Repo
import jinja2
from collections import OrderedDict
from templex import Templex
from pathquery import pathquery
from path import Path
import difflib
import mimetypes

CHANGELOG_MD_TEMPLATE = """\
# Changelog

{% for version, changes in version_changes.items() %}
### {% if version != None %}{{ version }}{% else %}Latest{% endif %}

{% for change in changes -%}
* {{ change }}
{%- else %}
No relevant code changes.
{%- endfor %}
{% endfor %}
"""

KEYWORDS = ["FEATURE", "BUGFIX", "BUG", "MINOR", "MAJOR", "PATCH", "PERFORMANCE"]


def changelog(project_dir):
    repo = Repo(project_dir)
    tag_commits = {tag.commit: tag for tag in repo.tags}
    current_version = Path(project_dir).joinpath("VERSION").text().rstrip()
    changes = []
    version_changes = OrderedDict()

    for commit in repo.iter_commits():
        if commit in tag_commits:
            version_changes[current_version] = changes
            current_version = tag_commits[commit].name
            changes = []

        message = commit.message

        for keyword in KEYWORDS:
            if message.startswith(keyword):
                if message not in changes:  # don't add dupes
                    changes.append(message)

    return (
        jinja2.Template(CHANGELOG_MD_TEMPLATE)
        .render(version_changes=version_changes)
        .replace("\r\n", "\n")
    )


def directory_template(all_stories, project_dir, story_dir, build_dir, readme=False):
    return (
        dirtemplate.DirTemplate(project_dir / "docs" / "src", build_dir)
        .with_files(
            template_story_jinja2={
                "using/alpha/{0}.md".format(story.info["docs"]): {"story": story}
                for story in all_stories.ordered_by_name()
                if story.info.get("docs") is not None
            }
        )
        .with_vars(
            readme=readme,
            quickstart=all_stories.in_filename(story_dir / "quickstart.story")
            .non_variations()
            .ordered_by_file(),
            include_title=True,
        )
        .with_functions(title=title)
    )


def title(dirfile):
    assert len(dirfile.text().split("---")) >= 3, "{} doesn't have ---".format(dirfile)
    return load(dirfile.text().split("---")[1]).data.get("title", "misc")


def docgen(all_stories, project_dir, story_dir, build_dir, temp_dir, check=False):
    """
    Generate markdown documentation.
    """

    docfolder = build_dir
    if check:
        if temp_dir.exists():
            temp_dir.rmtree(ignore_errors=True)
        temp_dir.mkdir()

        directory_template(
            all_stories, project_dir, story_dir, temp_dir, readme=False
        ).ensure_built()
        temp_dir.joinpath("changelog.md").write_text(changelog(project_dir))
        temp_dir.joinpath("fingerprint.txt").remove()
        print("Docs checked")

        assert len(list(pathquery(temp_dir))) == len(
            list(pathquery(docfolder))
        ), "Different real docs to generated"

        for temp_docfile in pathquery(temp_dir):
            if not temp_docfile.isdir():
                equivalent_realdocfile = Path(temp_docfile.replace(temp_dir, docfolder))
                print("Checking {}".format(equivalent_realdocfile))
                textfile = mimetypes.guess_type(temp_docfile)[
                    0
                ] is not None and mimetypes.guess_type(temp_docfile)[0].startswith(
                    "text"
                )

                if textfile:
                    error_message = "Generated file different from real,\n{}".format(
                        "".join(
                            difflib.ndiff(
                                equivalent_realdocfile.text().splitlines(1),
                                temp_docfile.text().splitlines(1),
                            )
                        )
                    )
                    assert (
                        equivalent_realdocfile.text() == temp_docfile.text()
                    ), error_message
                else:
                    assert (
                        equivalent_realdocfile.bytes() == temp_docfile.bytes()
                    ), "Generated file different from real, please rerun docgen or report bug."
    else:
        if docfolder.exists():
            docfolder.rmtree(ignore_errors=True)
        docfolder.mkdir()

        directory_template(
            all_stories, project_dir, story_dir, docfolder, readme=False
        ).ensure_built()
        docfolder.joinpath("changelog.md").write_text(changelog(project_dir))
        docfolder.joinpath("fingerprint.txt").remove()
        print("Docs generated")


def readmegen(
    all_stories, project_dir, story_dir, build_dir, project_name, check=False
):
    docfolder = build_dir / "readme"
    if docfolder.exists():
        docfolder.rmtree(ignore_errors=True)
    docfolder.mkdir()

    directory_template(
        all_stories, project_dir, story_dir, docfolder, readme=True
    ).ensure_built()

    import re

    text_with_absolute_links = re.sub(
        r"(\[.*?\])\(((?!http).*?)\)",
        r"\g<1>(https://hitchdev.com/{0}/\g<2>)".format(project_name),
        docfolder.joinpath("index.md").text(),
    ).replace("\r\n", "\n")

    if check:
        Templex(changelog(project_dir)).assert_match(
            project_dir.joinpath("CHANGELOG.md").read_text()
        )
        Templex(text_with_absolute_links).assert_match(
            project_dir.joinpath("README.md").read_text()
        )
        print("README and CHANGELOG are correct.")
    else:
        project_dir.joinpath("CHANGELOG.md").write_text(changelog(project_dir))
        project_dir.joinpath("README.md").write_text(text_with_absolute_links)
        print("README and CHANGELOG generated")



from hitchpylibrarytoolkit.deploy import deploy
from hitchpylibrarytoolkit.docgen import docgen
from hitchpylibrarytoolkit.docgen import readmegen
from hitchpylibrarytoolkit.exceptions import ToolkitError
from commandlib import python, python_bin, Command, CommandError
from hitchstory import StoryCollection
from pathquery import pathquery


class ProjectToolkit(object):
    def __init__(self, project_name, paths):
        self._project_name = project_name
        self._path = paths

    @property
    def current_version(self):
        return self._path.project.joinpath("VERSION").text().rstrip()

    def bdd(self, engine, keywords):
        """Run individual story matching key words."""
        self._stories(engine).only_uninherited().shortcut(*keywords).play()

    def regression(self, engine):
        self._stories(engine).only_uninherited().ordered_by_name().play()

    def _stories(self, engine):
        return StoryCollection(
            pathquery(self._path.key / "story").ext("story"),
            engine,
        )

    def prepdeploy(self, engine, version):
        """Prepare a deployment with new version, README and docs, but do not push."""
        git = Command("git").in_dir(self._path.project)
        version_file = self._path.project.joinpath("VERSION")
        old_version = version_file.text().rstrip()
        from packaging.version import parse

        if not parse(version) > parse():
            raise ToolkitError(
                "New version {} must be above old version {}".format(
                    version, old_version
                )
            )
        version_file.write_text(version)
        self.readmegen(engine)
        self.docgen(engine)
        self.validate_readmegen(engine)
        self.validate_docgen(engine)
        git("add", ".").run()
        git(
            "commit", "-m", "RELEASE: Version {0} -> {1}".format(old_version, version)
        ).run()
        git("tag", "-a", version, "-m", "Version {0}".format(version)).run()

    def pushdeploy(self):
        """Push a new deployment to pypi."""
        git = Command("git").in_dir(self._path.project)
        version = self._path.project.joinpath("VERSION").text().rstrip()
        git("push", "--follow-tags").run()

        # Set __version__ variable in __init__.py, build sdist and put it back
        initpy = self._path.project.joinpath(self._project_name, "__init__.py")
        original_initpy_contents = initpy.bytes().decode("utf8")
        initpy.write_text(
            original_initpy_contents.replace("DEVELOPMENT_VERSION", version)
        )
        python("setup.py", "sdist").in_dir(self._path.project).run()
        initpy.write_text(original_initpy_contents)

        # Upload to pypi
        python(
            "-m",
            "twine",
            "upload",
            "dist/{0}-{1}.tar.gz".format(self._project_name, version),
        ).in_dir(self._path.project).run()

    def deploy(self, version):
        deploy(self._path.project, self._project_name, version)

    def lint(self, exclude=None):
        try:
            if exclude is None:
                exclude = "__init__.py"
            python("-m", "flake8")(
                self._path.project.joinpath(self._project_name),
                "--max-line-length=100",
                "--exclude={}".format(",".join(exclude)),
                "--ignore=E203,W503",  # Ignore list[expression1 : expression2] failures
            ).run()
            python("-m", "flake8")(
                self._path.project.joinpath("hitch", "key.py"),
                "--max-line-length=100",
                "--ignore=E203,W503",
            ).run()
        except CommandError:
            raise ToolkitError("Linter failed")

    def reformat(self):
        python_bin.black(self._path.project / self._project_name).run()
        python_bin.black(self._path.project / "hitch" / "key.py").run()

    def validate_reformatting(self):
        try:
            python_bin.black(self._path.project / self._project_name, "--check").run()
            python_bin.black(self._path.project / "hitch" / "key.py", "--check").run()
        except CommandError:
            raise ToolkitError("Black failed")

    def docgen(self, engine):
        docgen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.project / "docs" / "public",
            self._path.gen / "tmpdocs",
        )

    def validate_docgen(self, engine):
        docgen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.project / "docs" / "public",
            self._path.gen / "tmpdocs",
            check=True,
        )

    def readmegen(self, engine):
        readmegen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.gen,
            self._project_name,
        )

    def validate_readmegen(self, engine):
        readmegen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.gen,
            self._project_name,
            check=True,
        )



from commandlib import CommandPath, Command
from distutils.version import LooseVersion, StrictVersion
from path import Path
import hitchbuild
import requests
import dateutil
from typing import List
import tomli


def package_versions_above(package_name, minimum_version):
    data = requests.get("https://pypi.org/pypi/{}/json".format(package_name)).json()
    versions = list(
        [release for release in data["releases"].keys() if "rc" not in release]
    )
    versions.sort(key=StrictVersion)

    selected_versions = [
        version
        for version in versions
        if LooseVersion(version) >= LooseVersion(minimum_version)
    ]

    upload_dates = {
        version: dateutil.parser.parse(data["releases"][version][-1]["upload_time"])
        for version in selected_versions
    }

    possible_versions = []

    for index, version in enumerate(selected_versions[1:]):
        timediff = upload_dates[version] - upload_dates[selected_versions[index]]

        if timediff.days > 7:
            possible_versions.append(selected_versions[index])

    possible_versions.append(selected_versions[-1])

    return possible_versions


class ProjectDependencies:
    def __init__(self, pyproject_toml, pyenv_build):
        self._pyproject_toml = pyproject_toml
        self._pyenv_build = pyenv_build

    def load(self):
        project = tomli.loads(self._pyproject_toml)["project"]
        pyversion = project["requires-python"]
        dependencies = project["dependencies"]

        self.dependency_versions = {}

        for dependency in dependencies:
            assert ">=" in dependency

            self.dependency_versions[
                dependency.split(">=")[0]
            ] = package_versions_above(
                dependency.split(">=")[0],
                dependency.split(">=")[1],
            )

        self.python_versions = self._pyenv_build.available_versions_above_and_including(
            pyversion.split(">=")[1]
        )


class EnvirotestVirtualenv(hitchbuild.HitchBuild):
    def __init__(
        self, pyenv_build, pyproject_toml, picker, test_package, prerequisites
    ):
        self._pyenv_build = pyenv_build
        self._pyproject_toml = pyproject_toml
        self._picker = picker
        self._test_package = test_package
        self._prerequisites = prerequisites

    def build(self):
        self._pyenv_build.ensure_built()

        project_dependencies = ProjectDependencies(
            self._pyproject_toml,
            self._pyenv_build,
        )
        project_dependencies.load()

        self.python_version = self._picker(project_dependencies.python_versions[:-1])
        print("Python version: {}".format(self.python_version))

        self.picked_versions = {}
        for dependency, versions in project_dependencies.dependency_versions.items():
            self.picked_versions[dependency] = self._picker(versions)
            print("{} version: {}".format(dependency, self.picked_versions[dependency]))

        self.venv = ProjectVirtualenv(
            "randomtestvenv",
            PyVersion(
                self._pyenv_build,
                self.python_version,
            ),
            packages=self._prerequisites
            + [
                PythonRequirements(
                    [
                        "{}=={}".format(library, version)
                        for library, version in self.picked_versions.items()
                    ]
                ),
                self._test_package,
            ],
        )
        self.venv.clean()
        self.venv.ensure_built()


class DevelopmentVirtualenv(hitchbuild.HitchBuild):
    def __init__(
        self,
        pyenv_build,
        versions_file,
        debug_requirements,
        project_path,
        pyproject_toml,
    ):
        self._pyenv_build = pyenv_build
        self._versions_file = versions_file
        self._debug_requirements = debug_requirements
        self._project_path = project_path
        self._pyproject_toml = pyproject_toml

    @property
    def build_path(self):
        return self._pyenv_build.build_path / "versions" / "devvenv"

    @property
    def fingerprint_path(self):
        return self.build_path / "fingerprint.txt"

    @property
    def python_path(self):
        return self.build_path / "bin" / "python"

    def build(self):
        self._pyenv_build.ensure_built()

        if not self.build_path.exists():
            if not self._versions_file.exists():
                project_dependencies = ProjectDependencies(
                    self._pyproject_toml,
                    self._pyenv_build,
                )
                project_dependencies.load()

                self.python_version = project_dependencies.python_versions[:-1][-1]

                self.picked_versions = {}
                for (
                    dependency,
                    versions,
                ) in project_dependencies.dependency_versions.items():
                    self.picked_versions[dependency] = versions[-1]
            else:
                from strictyaml import load

                devenv = load(self._versions_file.text()).data
                self.python_version = devenv["python version"]
                self.picked_versions = devenv["packages"]

            self.venv = ProjectVirtualenv(
                "devvenv",
                PyVersion(
                    self._pyenv_build,
                    self.python_version,
                ),
                packages=[
                    PythonRequirements(
                        ["ensure", "python-slugify"],
                    ),
                    PythonRequirements(
                        [
                            "{}=={}".format(library, version)
                            for library, version in self.picked_versions.items()
                        ]
                    ),
                    PythonProjectDirectory(self._project_path),
                ],
            )
            self.venv.ensure_built()
            self.refingerprint()


class PythonPackage:
    pass


class PythonVersionDependentRequirement(PythonPackage):
    def __init__(
        self, package, lower_version, python_version_threshold, higher_version
    ):
        self._package = package
        self._lower_version = lower_version
        self._python_version_threshold = LooseVersion(python_version_threshold)
        self._higher_version = higher_version

    def install(self, venv):
        if LooseVersion(venv.py_version.version) > self._python_version_threshold:
            venv.pip("install", "{}={}".format(self._package, self._higher_version))
        else:
            venv.pip("install", "{}={}".format(self._package, self._lower_version))


class PythonRequirementsFile(PythonPackage):
    def __init__(self, requirements_file):
        self.requirements_file = requirements_file

    def install(self, venv):
        venv.pip("install", "-r", self.requirements_file).run()


class PythonRequirements(PythonPackage):
    def __init__(self, requirements: List[str], test_repo=False):
        self.requirements = requirements
        self._test_repo = test_repo

    def install(self, venv):
        for requirement in self.requirements:
            if self._test_repo:
                venv.pip(
                    "install",
                    "--no-build-isolation",
                    "--index-url",
                    "https://test.pypi.org/simple/",
                    requirement,
                ).run()
            else:
                venv.pip("install", requirement).run()


class PythonProjectPackage(PythonPackage):
    def __init__(self, package_filename):
        self.package_filename = package_filename

    def install(self, venv):
        local_install_output = venv.pip("install", self.package_filename).output()

        if "DEPRECATION" in local_install_output:
            raise Exception("DEPRECATION ERROR:\n {}".format(local_install_output))


class PythonProjectDirectory(PythonPackage):
    def __init__(self, project_directory):
        self.project_directory = project_directory

    def install(self, venv):
        venv.pip("install", "-e", self.project_directory).run()


class ProjectVirtualenv(hitchbuild.HitchBuild):
    def __init__(
        self,
        venv_name,
        py_version,
        requirements=None,
        requirements_files=None,
        package_dir=None,
        local_package=None,
        packages=None,
    ):
        self.venv_name = venv_name
        self.py_version = py_version
        self.packages = packages
        self.build_path = (
            self.py_version.pyenv_build.build_path / "versions" / self.venv_name
        )
        self.fingerprint_path = self.build_path / "fingerprint.txt"

    @property
    def pyenv(self):
        return self.py_version.pyenv_build.pyenv

    @property
    def python_path(self):
        return self.build_path / "bin" / "python"

    def clean(self):
        try:
            self.build_path.readlink().rmtree(ignore_errors=True)
            self.build_path.remove()
        except FileNotFoundError:
            pass

    @property
    def pip(self):
        return Command(self.build_path / "bin" / "pip")

    def build(self):
        if not self.build_path.exists():
            self.py_version.ensure_built()
            self.pyenv("virtualenv", self.py_version.version, self.venv_name).run()

            if self.packages is not None:
                for package in self.packages:
                    package.install(self)

            self.refingerprint()


class PyVersion(hitchbuild.HitchBuild):
    def __init__(self, pyenv_build, version):
        self.pyenv_build = pyenv_build
        self.version = version
        self.build_path = self.pyenv_build.build_path / "versions" / self.version
        self.fingerprint_path = self.build_path / "fingerprint.txt"

    def clean(self):
        self.build_path.rmtree(ignore_errors=True)

    def build(self):
        if not self.build_path.exists():
            self.pyenv_build.ensure_built()
            self.pyenv_build.pyenv("install", self.version).run()
            self.refingerprint()


class Pyenv(hitchbuild.HitchBuild):
    def __init__(self, build_path):
        self.build_path = Path(build_path).abspath()
        self.fingerprint_path = self.build_path / "fingerprint.txt"

    @property
    def bin(self):
        return CommandPath(self.build_path / "bin")

    def clean(self):
        self.build_path.rmtree(ignore_errors=True)

    @property
    def pyenv(self):
        return Command(self.bin.pyenv).with_env(PYENV_ROOT=self.build_path)

    def available_versions_above_and_including(self, minimum_version):
        return [
            version.strip()
            for version in self.pyenv("install", "-l").output().split("\n")
            if version.strip().startswith("3")
            and "dev" not in version
            and LooseVersion(version.strip()) >= LooseVersion(minimum_version)
        ]

    def build(self):
        if self.incomplete():
            self.build_path.rmtree(ignore_errors=True)
            self.build_path.mkdir()
            Command(
                "git", "clone", "https://github.com/pyenv/pyenv.git", self.build_path
            ).run()
            Command(
                "git",
                "clone",
                "https://github.com/pyenv/pyenv-virtualenv.git",
                self.build_path / "plugins" / "pyenv-virtualenv",
            ).run()
            self.refingerprint()
